MODULE DE

  USE global
  USE dice_objfn
  USE rndseed

  implicit none

#ifdef HAVE_MPI
!  include "mpif.h"
#endif

  integer, parameter :: refresh=100
  integer, parameter :: interval = 99
  integer, parameter :: iwrite=7
  integer :: strategy 
  integer, dimension(2) :: method
  double precision, parameter :: VTR=-100000000d0, CR_XC=0.5d0
  double precision, parameter :: DE_pen_val = 10000.0d0
  integer :: max_iters
  integer :: NP
  integer, pointer :: n_x
  integer, pointer :: n_g
  double precision :: in_popf   ! factor to multiply n_x to get NP 
  integer :: in_rndscale        ! method(1)
  integer :: in_strat           ! strategy 1 to 6
  double precision :: in_F_CR   ! Crossover factor: 0.0 implies random
                                ! and method(2) = 1.  Otherwise method(2) = 0
  double precision, dimension(:), allocatable :: x_lower, x_upper
  integer :: nfeval

CONTAINS
!---------------------------------------------------------------------------
SUBROUTINE run_DE()
!  =====================================================================
! |  This routine is tailored for use with DICE and replaces the        |
! |  original DE driver subroutine called Rosen which is included in    |
! |  comments below.                                                    |
! |  Brian Tuttle 15Oct08 <btuttle@psu.edu>                             |
!  =====================================================================

    USE mt19937

    implicit none

    double precision, dimension(:), allocatable :: solution
    integer :: seed
    double precision :: objf
    integer :: astat, istat
    integer :: UID = 20
    integer(i4b) :: stat = 0
    
! Read input file.
    call read_input(stat)
    if (stat > 0) then
        print *, "run_DE: Error on read_input."
        RETURN
    end if

! Verify by printing what was just read.
    if (iam == 0) then
        call print_input()
    end if

! Get the size of the model.
    call model_size(stat)
    if (stat > 0) then
        print *, "run_DE: Error on dice_optim_size."
        RETURN
    end if

! Assign size pointers.
    n_x => nop
    n_g => ng

! Allocate upper and lower bounds.
    astat = 0
    allocate(x_lower(n_x), STAT=astat)
    allocate(x_upper(n_x), STAT=astat)
    allocate(solution(n_x), STAT=astat)
    if (astat > 0) then
        print *, "Error allocating x_lower, x_upper."
        RETURN
    end if

! Initialize model.
    call model_setup(x_lower, x_upper, stat)
    if (stat > 0) then
        print *, "run_DE: Error on model_setup."
        RETURN
    end if

    NP = nint(in_popf * n_x)
    method(1) = in_rndscale
    method(2) = 0
    if (in_F_CR == 0.0d0) method(2) = 1
    strategy = in_strat


! Initialize random seed.
        if (iam == 0) then
          seed = randate()
        end if
#ifdef HAVE_MPI
        call MPI_Bcast(seed, 1, MPI_Integer4, 0, MPI_Comm_World, ierr)
        seed = mod(seed + iam, max_rnd_seed)
#endif
        print *, "processor", iam, "seed = ", seed
        call init_genrand(seed)

! Open the progress report file and write a header to it.
    if (iam == 0) then
        open (UNIT=UID, FILE='fout', STATUS='REPLACE', ACCESS='SEQUENTIAL', &
            IOSTAT=istat)
        write(UID,"(a)") "    Iteration     Objf evals    Objf value"

        close(UID)
    end if

! Do the optimization.
    call DE_Fortran90(max_iters, solution, in_F_CR)

! Report the final solution.
    if (iam == 0) then
        call model_out(solution, objf)
        print *, "Final objective function value: ", objf
    end if

! Clean up.
    nullify(n_x)
    nullify(n_g)

    astat = 0
    deallocate(x_lower, STAT=astat)
    deallocate(x_upper, STAT=astat)
    deallocate(solution, STAT=astat)
    if (astat > 0) print *, "Error deallocating x_lower, x_upper."

    call dealloc_dice_objf()

    RETURN

END SUBROUTINE run_DE
!---------------------------------------------------------------------------

!---------------------------------------------------------------------------
SUBROUTINE DE_Fortran90(miter, bestmem, CRfact)
!=======================================================================
!  
!   Differential Evolution for Optimal Control Problems
!
!-----------------------------------------------------------------------
!    This Fortran 90 program translates from the original MATLAB 
!    version of differential evolution (DE). This FORTRAN 90 code 
!    has been tested on Compaq Visual Fortran v6.1. 
!    Any users new to the DE are encouraged to read the article of 
!    Storn and Price. 
!
!  Refences:
!    Storn, R., and Price, K.V., (1996). Minimizing the double precision 
!    function of the ICEC96 contest by differential evolution. IEEE conf. 
!    on Evolutionary Comutation, 842-844.
!
!  This Fortran 90 program written by Dr. Feng-Sheng Wang 
!  Department of Chemical Engineering, National Chung Cheng University, 
!  Chia-Yi 621, Taiwan, e-mail: chmfsw@ccunix.ccu.edu.tw
!=======================================================================
!             obj : The user provided file for evlauting the objective function.
!                   subroutine obj(xc,fitness)
!                   where "xc" is the double precision decision parameter 
!                   vector.(input)
!                   "fitness" is the fitness value.(output)
!             n_x : Dimension of the double precision decision parameters.
!    x_lower(n_x) : The lower bound of the double precision decision parameters.
!    x_upper(n_x) : The upper bound of the double precision decision parameters.
!             VTR : The expected fitness value to reach.
!              NP : Population size.
!         itermax : The maximum number of iteration.
!            F_XC : Mutation scaling factor for double precision decision 
!                   parameters.
!           CR_XC : Crossover factor for double precision decision parameters.
!        strategy : The strategy of the mutation operations is used in HDE.
!         refresh : The intermediate output will be produced after "refresh"
!                   iterations. No intermediate output will be produced if
!                   "refresh < 1".
!        interval : The intermediate output will be produced after "interval"
!                   iterations following the last update to bestval.
!          iwrite : The unit specfier for writing to an external data file.
! bestmem_XC(n_x) : The best double precision decision parameters.
!         bestval : The best objective function.
!          nfeval : The number of function call.
!         method(1) = 0, Fixed mutation scaling factors (F_XC)
!                   = 1, Random mutation scaling factors F_XC=[0, 1]
!                   = 2, Random mutation scaling factors F_XC=[-1, 1] 
!         method(2) = 1, Random combined factor (F_CR) used for strategy = 6
!                        in the mutation operation 
!                   = other, fixed combined factor provided by the user 
!         method(3) = 1, Saving results in a data file.
!                   = other, displaying results only.
! ========================================================================
    
    USE mt19937

    implicit none

    integer, intent(INOUT) :: miter
    double precision, dimension(:), intent(OUT) :: bestmem
    double precision, intent(IN) :: CRfact
    double precision, dimension(NP,n_x) :: pop_XC, bm_XC, mui_XC, &
         mpo_XC, popold_XC, rand_XC, ui_XC
    double precision :: F_XC, F_CR
    integer, dimension(NP) :: rot, rt
    integer, dimension(NP) :: a1, a2, a3, a4, a5
    integer, dimension(4) :: ind
    double precision, dimension(NP) :: tempval, tempval_pr, phi, phi_pr
    double precision, dimension(NP) :: val
    double precision :: bestval
    double precision, dimension(n_x) :: bestmem_XC
    intrinsic max, min, random_number, mod, abs, any, all, maxloc
    double precision :: fbest, phibest
    logical :: feasible
    integer :: i, ibest, total_iters, iter, nfeval_sum
    integer(i8b) :: nfe_total

! Initialize mutation factors.
    F_XC = 0.8d0
    F_CR = CRfact

! Initialize the population and evaluate fitness functions to find the best 
! member.
    pop_XC=0.0d0
    val=0.0d0
    bestval = 0.0d0

    nfe_total = 0
    nfeval = 0

    call init_pop(pop_XC(:,:), bestmem_XC(:), val(:), bestval, feasible)
 
    if (iam == 0) then
        bm_XC=0.0d0
        rot=(/(i,i=0,NP-1)/)
    end if

! Perform evolutionary computation: Start TIME LOOP.

! timeloop: do
  do iter = 1, miter

    if (iam == 0) then      ! MASTER processor task
       popold_XC=pop_XC

! Mutation operation:
        ind=randperm(4)
        a1(:)=randperm(NP)
        rt=mod(rot+ind(1),NP)
        a2(:)=a1(rt+1)
        rt=mod(rot+ind(2),NP)
        a3(:)=a2(rt+1)
        rt=mod(rot+ind(3),NP)
        a4(:)=a3(rt+1)
        rt=mod(rot+ind(4),NP)
        a5(:)=a4(rt+1)
       bm_XC=spread(bestmem_XC, DIM=1, NCOPIES=NP)

! Generating a random scaling factor
       select case (method(1))
       case (1)
          F_XC = genrand_real1()
       case(2)
          F_XC = genrand_real1()
          F_XC=2.0d0*F_XC-1.0d0
       end select

! Select a mutation strategy.
       select case (strategy)
       case (1)
           ui_XC(:,:) = bm_XC(:,:) + &
                     F_XC*(popold_XC(a1(:),:)-popold_XC(a2(:),:))
          
       case default
          ui_XC(:,:) = popold_XC(a3(:),:) + &
                    F_XC*(popold_XC(a1(:),:)-popold_XC(a2(:),:))
          
       case (3)
          ui_XC(:,:) = popold_XC(:,:) + &
                    F_XC*(bm_XC(:,:)-popold_XC(:,:) + &
                             popold_XC(a1(:),:)-popold_XC(a2(:),:))
          
       case (4)
          ui_XC(:,:) = bm_XC(:,:) + &
                  F_XC*(popold_XC(a1(:),:)-popold_XC(a2(:),:) + &
                           popold_XC(a3(:),:)-popold_XC(a4(:),:))
          
       case (5)
          ui_XC(:,:) = popold_XC(a5(:),:) + &
                  F_XC*(popold_XC(a1(:),:)-popold_XC(a2(:),:) + &
                           popold_XC(a3(:),:)-popold_XC(a4(:),:))
       case (6) ! A linear crossover combination of bm_XC and popold_XC
          if (method(2) == 1) F_CR = genrand_real1()
          ui_XC(:,:) = popold_XC(:,:) + &
                    F_CR*(bm_XC(:,:)-popold_XC(:,:)) + &
                    F_XC*(popold_XC(a1(:),:)-popold_XC(a2(:),:))
          
       end select

! Crossover operation:
       call rnd_dist(rand_XC)
       mui_XC=0.0d0
       mpo_XC=0.0d0
       where (rand_XC < CR_XC)
          mui_XC=1.0d0
       elsewhere
          mpo_XC=1.0d0
       end where

       ui_XC=popold_XC*mpo_XC+ui_XC*mui_XC

       do i=1,NP
! Confine each of feasible individuals in the lower-upper bound.
         ui_XC(i,:)=max(min(ui_XC(i,:),x_upper),x_lower)
       end do

    end if      !! End MASTER task

#ifdef HAVE_MPI
    call MPI_Bcast(ui_XC, NP*n_x, MPI_Double_Precision, 0, &
                    MPI_Comm_World, ierr)
#endif


         tempval_pr = 0.0d0
         phi_pr = 0.0d0
         do i=1+iam,NP,n_procs
           call objfunc(ui_XC(i,:), tempval_pr(i), phi_pr(i))
         end do

#ifdef HAVE_MPI
         Call MPI_Reduce(tempval_pr, tempval, NP, MPI_Double_Precision, &
                        MPI_Sum, 0, MPI_Comm_World, ierr)
         Call MPI_Reduce(phi_pr, phi, NP, MPI_Double_Precision, &
                        MPI_Sum, 0, MPI_Comm_World, ierr)
#else
         tempval = tempval_pr
         phi = phi_pr
#endif

         if (iam == 0) then     !** MASTER task

           do i=1,NP
             if (tempval(i) < val(i)) then ! .and. phi(i) == 0.0d0) then
               pop_XC(i,:)=ui_XC(i,:)
               val(i) = tempval(i)

               if (tempval(i) < bestval) then !  .and. phi(i) == 0.0d0) then
                 bestval = tempval(i)
                 bestmem_XC(:) = ui_XC(i,:)
                 if (phi(i) == 0.0d0) then
                    feasible = .true.
                 else
                    feasible = .false.
                 end if
               end if

             end if
           end do

         end if                   ! end MASTER task **

       if( (refresh > 0) .and. (mod(iter,refresh)==0)) then
#ifdef HAVE_MPI
          call MPI_Reduce(nfeval, nfeval_sum, 1, MPI_Integer, MPI_SUM, 0, &
                            MPI_COMM_WORLD, ierr)
          nfe_total = nfe_total + nfeval_sum
#else
          nfe_total = nfe_total + nfeval
#endif
          nfeval = 0

          if (iam == 0) then
              call write_progress(iter, nfe_total, bestval)
          end if

       end if
            
  end do      ! End miter loop.

#ifdef HAVE_MPI
    call MPI_Bcast(bestmem_XC, n_x, MPI_Double_Precision, 0, &
                   MPI_Comm_World, ierr)
#endif
    call objfunc(bestmem_XC(:), fbest, phibest)

    if (iam == 0 .and. phibest > 0.0d0 ) then
       print *, "Solution is not feasible."
       print *, "Increase pen_val in subroutine objfunc in DE.f90" 
    end if

    bestmem = bestmem_XC

    RETURN

END SUBROUTINE DE_Fortran90
!---------------------------------------------------------------------------

!---------------------------------------------------------------------------
SUBROUTINE init_pop(poparray, best_mem, obfval, best_obfval, feasible_soln)
!  =======================================================================
! |  Initializes the population with a random array, evaluates each       |
! |  member via the objective function and finds the best objective       |
! |  function value along with the best solution.                         |
!  =======================================================================

    implicit none

    double precision, dimension(NP,n_x), intent(OUT) :: poparray
    double precision, dimension(n_x), intent(OUT) :: best_mem
    double precision, dimension(NP), intent(OUT) :: obfval
    double precision, intent(OUT) :: best_obfval
    logical, intent(OUT) :: feasible_soln
    double precision, dimension(NP) :: obfval_pr, phi_pr, phi
    double precision, dimension(n_x) :: rand_C1
    integer :: i, ibest, j

! Initialize a random population.
    if (iam == 0) then
        poparray = 0.0d0
        do i=1,NP
            call rnd_dist(rand_C1)
            poparray(i,:) = x_lower + rand_C1*(x_upper-x_lower)
        end do
    end if

    feasible_soln = .true.
#ifdef HAVE_MPI
! Distribute population array to the rest of the processors.
    call MPI_Bcast(poparray, NP*n_x, MPI_Double_Precision, 0, &
                    MPI_Comm_World, ierr)
    call MPI_Barrier(MPI_Comm_World, ierr)
#endif

! Evaluate fitness function to find the best member.
    obfval = 0.0d0
    obfval_pr = 0.0d0
    phi = 0.0d0
    phi_pr = 0.0d0
    do i=1+iam,NP,n_procs
        call objfunc(poparray(i,:), obfval_pr(i), phi_pr(i))
    end do

#ifdef HAVE_MPI
    call MPI_Reduce(obfval_pr, obfval, NP, MPI_Double_Precision, &
                    MPI_Sum, 0, MPI_Comm_World, ierr)
    call MPI_Reduce(phi_pr, phi, NP, MPI_Double_Precision, &
                    MPI_Sum, 0, MPI_Comm_World, ierr)
#else
    obfval = obfval_pr
    phi = phi_pr
#endif

    if (iam == 0) then
        ibest=1
        best_obfval = obfval(1)
        do i=2,NP
          if (obfval(i)<best_obfval) then ! .and. phi(i) == 0.0d0) then
            ibest = i
            best_obfval = obfval(i)
            if (phi(i) > 0.0d0) then
               feasible_soln = .false.
            else
               feasible_soln = .true.
            end if
          end if
        end do

        best_mem = poparray(ibest,:)
    end if

#ifdef HAVE_MPI
    call MPI_Bcast(feasible_soln, 1, MPI_Logical, 0, MPI_Comm_World, ierr)
#endif

    RETURN

END SUBROUTINE init_pop
!---------------------------------------------------------------------------
       
!---------------------------------------------------------------------------
FUNCTION randperm(num)
!  =======================================================================
! |  Returns a random permutation of numbers from 1 to num.               |
!  =======================================================================

    implicit none

    integer, intent(in) :: num
    integer :: number, i, j, k
    integer, dimension(num) :: randperm
    double precision, dimension(num) :: rand2
    intrinsic random_number

    call rnd_dist(rand2)
    do i=1,num
       number=1
       do j=1,num
          if (rand2(i) > rand2(j)) then
             number=number+1
          end if
       end do
       do k=1,i-1
          if (rand2(i) <= rand2(k) .and. rand2(i) >= rand2(k)) then
             number=number+1
          end if
       end do
       randperm(i)=number
    end do

    RETURN

END FUNCTION randperm
!---------------------------------------------------------------------------

!---------------------------------------------------------------------------
SUBROUTINE objfunc(X, objval, phi)
!  =======================================================================
! |  This subroutine handles penalties for DE.  It is the generic         |
! |  objective function routine which calls the specific model function.  |
!  =======================================================================

    implicit none

    double precision, dimension(n_x), intent(IN) :: X
    double precision, intent(OUT) :: objval, phi
    double precision :: f
    double precision, dimension(n_g) :: g
    integer :: i
    integer :: nfe     ! placeholder for number of function evaluations

    phi = 0.0d0

! Default number of function evaluations is 1.
    nfe = 1

    call calc_fg_dice(X,f,g,nfe)

! Lower objval is better, so penalties are positive when added.
    objval = f
    do i=1,n_g
       if (g(i).gt.0d0) then
        objval = objval + DE_pen_val*g(i)
        phi = phi + g(i)
       end if
    end do

    nfeval = nfeval + nfe
    
END SUBROUTINE objfunc
!---------------------------------------------------------------------------

!---------------------------------------------------------------------------
SUBROUTINE read_input(errcode)
!  ========================================================================
! |  This subroutine reads from a file called input, assigns values to     |
! |  the input variables, and broadcasts the input data to the rest of the |
! |  processors if any.                                                    |
!  ========================================================================


    implicit none

    integer(i4b), intent(OUT) :: errcode
    integer, parameter :: tabstop = 31
    integer, parameter :: IOU = 10
    character(len=tabstop) :: tabover
    character(len=75) :: sec_head
    integer :: istat, i
    logical :: archivefile
    integer(i4b) :: stat = 0

    errcode = 0

if (iam == MASTER) then

    ! Open input file.
        open(UNIT=IOU, FILE='input', STATUS='OLD', ACTION='READ', &
             FORM='FORMATTED', IOSTAT=istat)

  headers: do
    read(IOU, "(a)", IOSTAT=istat) sec_head

    select case (sec_head(1:tabstop))
    case ('--------------------------- DE ')
      de_param: do
        read(IOU, "(a)", ADVANCE="NO", IOSTAT=istat) tabover
        select case (tabover)
        case ('  maximum iterations         = ')
          read(IOU, "(i12)", IOSTAT=istat) max_iters
        case ('  population factor          = ')
           read(IOU, "(f7.4)", IOSTAT=istat) in_popf
        case ('  Random scaling factor      = ')
           read(IOU, "(i4)", IOSTAT=istat) in_rndscale
        case ('  mutation strategy [1-6]    = ')
           read(IOU, "(i4)", IOSTAT=istat) in_strat
        case ('  crossover factor [0.8]     = ')
           read(IOU, "(f7.4)", IOSTAT=istat) in_F_CR
        case ('                               ')
          exit de_param
        end select
      end do de_param
    case ('------------------------ MODEL ')
      exit headers
    end select
  end do headers

end if

! --- Model parameters
    call read_dice_input(IOU, stat)
    errcode = errcode + stat

if (iam == MASTER) then

! Close input file.
    close(IOU)

end if

#ifdef HAVE_MPI
    call MPI_Bcast(max_iters, 1, MPI_Integer, MASTER, MPI_COMM_WORLD, ierr)
    call MPI_Bcast(in_popf, 1, MPI_Double_Precision, MASTER, MPI_COMM_WORLD, ierr)
    call MPI_Bcast(in_rndscale, 1, MPI_Integer, MASTER, MPI_COMM_WORLD, ierr)
    call MPI_Bcast(in_strat, 1, MPI_Integer, MASTER, MPI_COMM_WORLD, ierr)
    call MPI_Bcast(in_F_CR, 1, MPI_Double_Precision, MASTER, MPI_COMM_WORLD, ierr)
#endif

    RETURN

END SUBROUTINE read_input
!---------------------------------------------------------------------------

!---------------------------------------------------------------------------
SUBROUTINE print_input()
!  =======================================================================
! |  Print out what was read from the input file.                         |
!  =======================================================================

    implicit none

    print *, "DE PARAMETERS:"
    print *, "    maximum iterations       = ", max_iters
    print *, "    population factor        = ", in_popf
    print *, "    Random scaling factor    = ", in_rndscale
    print *, "    mutation strategy [1-6]  = ", in_strat
    print *, "    crossover factor [0.8]   = ", in_F_CR
    print *, ""

    call print_model_input()

    RETURN

END SUBROUTINE print_input
!---------------------------------------------------------------------------

!---------------------------------------------------------------------------
SUBROUTINE write_progress(iteration, obf_evals, obf)
!  =======================================================================
! |  This routine updates the objective function output file with the     |
! |  iteration number, number of objective function evaluations and the   |
! |  value returned by the objective function.                            |
!  =======================================================================

    implicit none

    integer, intent(IN) :: iteration
    integer(i8b), intent(IN) :: obf_evals
    double precision, intent(IN) :: obf
    integer :: istat
    integer :: fileID = 20

    open(UNIT=fileID, FILE='fout', STATUS='OLD', ACCESS='SEQUENTIAL', &
            ACTION='WRITE', POSITION='APPEND', IOSTAT=istat)

    write(fileID,"(i12,i16,F21.11)") iteration, obf_evals, obf

    close(fileID)

    RETURN

END SUBROUTINE write_progress
!---------------------------------------------------------------------------
!!********************************************************************
!
!  subroutine Rosen
!
!!********************************************************************
!
!    implicit none
!    double precision, allocatable :: x(:)
!    double precision :: f
!    integer :: i
!    integer , dimension(8) :: time
!    integer :: alloc_stat, istat
!    integer :: seed(2)
!    integer :: UID = 20
!
!    intrinsic date_and_time
!    open(iwrite,file='Rosen.txt')
!    call date_and_time(values=time)
!    write(unit=iwrite, FMT=11) time(1:3), time(5:7)
!
!    call setup_go_problem
!
!    allocate(x(n_x))
!
!       seed(1) = randate()
!!       seed(1) = 1 
!       seed(2) = seed(1)**0.5
!       print*,"seed =",seed
!
! call random_seed(put=seed)
!
!    alloc_stat = 0
!    allocate(bestmem_XC(n_x),stat=alloc_stat)
!    if (alloc_stat > 0) then
!       print *, "Error allocating memory for bestmem_XC"
!       stop
!    end if
!
!! Open the progress report file and write a header to it.
!    open (UNIT=UID, FILE='fout', STATUS='REPLACE', ACCESS='SEQUENTIAL', &
!            IOSTAT=istat)
!    write(UID,"(a)") "Iteration     Objf evals    Objf value"
!
!    close(UID)
!
! print*,"NP =",NP
! print*,"itermax =",itermax
! print*,"method(1:3) =",method(1:3)
! write(*,201) F_XC, CR_XC, F_CR
!
!    call DE_Fortran90
!
!    call write_output()
!
!    write(iwrite,205) NP, nfeval, method(1:3)
!    write(iwrite,FMT=201) F_XC, CR_XC, F_CR
!    write(iwrite,FMT=200) bestval
!    do i=1,n_x
!       write(iwrite,FMT=202) i,bestmem_XC(i)
!    end do
! 200 format(/2x, 'Bestval=', ES14.7)
! 201 format(2x, 'F_XC =',F6.3, 2x, 'CR_XC =', F6.3, 2x, 'F_CR =', F6.3)
! 202 format(2x, 'best_XC(',I3,') =',ES14.7)
! 205 format(2x, 'NP=', I4, 4x, 'No. function call =', I9, &
!         /2x, 'method(1:3) =',3I3)
!    call date_and_time(values=time)
!    write(unit=iwrite, FMT=10)time(1:3), time(5:7)
! 10  format(/1x, 'End of Program. Date:', I4, '/', I2,'/', I2, ', Time: ', I2,':',I2,':',I2)
! 11  format(1x, 'Beginning of Program. Date:', I4, '/', I2,'/', I2, ', Time: ', I2,':',I2,':',I2)
!
!  end subroutine Rosen
!
!!********************************************************************
END MODULE DE
