!    DICE: DICE07 with variable time step
!    Copyright (C) 2010  Klaus Keller and Brian Tuttle 
!
!    This program is free software; you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation; either version 2 of the License, or
!    (at your option) any later version.
!
!    This program is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with this program; if not, write to the Free Software
!    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
!
!    Klaus Keller, klaus.psu@gmail.com
!    Brian Tuttle, btuttle@psu.edu
!

!-----------------------------------------------------------------------------
MODULE DICE

    USE global
    USE dice_geo

    implicit none

    integer(i4b) :: ns, n_hist
    integer(i4b) :: n_geophys

    logical :: damages_on

! Exogenously defined variables:
    real(SP), dimension(:), allocatable :: time_yrs
    real(DP), dimension(:), allocatable :: RR, L, GA, AL, PART_FRAC, SIGMA
    real(DP), dimension(:), allocatable :: COST_1, E_TREE, FORCOTH
    real(DP), dimension(:), allocatable :: R    ! v7 GAMS

! Climate damage parameters:
    real(DP) :: AA1, AA2, AA3, theta3, theta4

! Economic parameters:
    real(DP) :: ELAS_MU, GAMA, K0, DK 
    real(DP) :: b2          ! Exponent of control cost function
    real(DP) :: DR      ! Discount rate v7 GAMS

! Endogenously defined variables:
    real(DP), dimension(:), allocatable :: b1, frac_abate_cost, frac_damages
    real(DP), dimension(:), allocatable :: K, YGROSS, Y, production
    real(DP), dimension(:), allocatable :: E, I, C, PERIOD_U

! Objective function:
    real(DP) :: scale1, scale2

CONTAINS
!-----------------------------------------------------------------------------
SUBROUTINE dice_size()
!  ==========================================================================
! |  Determine the size of DICE model arrays and allocate them.              |
!  ==========================================================================

    USE dice_in

    implicit none

    ns = int((end_year_in - start_year_in)/dt_in) + 1

    n_hist = 0

! Model arrays:
    call init_DICEgeo_arrays(ns, dt_in)

    call alloc_dice()

    RETURN

END SUBROUTINE dice_size
!-----------------------------------------------------------------------------

!-----------------------------------------------------------------------------
SUBROUTINE init_dice()
!  ==========================================================================
! |  Steps to initialize dice model.
!  ==========================================================================

    implicit none

! Exogenous variables:
    call dice_exog_var()

! Define fixed model parameters:
    call dice_params()

    call dice_geophys_params()

    RETURN

END SUBROUTINE init_dice
!-----------------------------------------------------------------------------

!-----------------------------------------------------------------------------
SUBROUTINE dice_params()
!  ==========================================================================
! |  Define exogenous variables and fixed parameters.                        |
!  ==========================================================================

    implicit none

! Fixed parameters:
! Climate damage parameters:
    AA1 = 0.0d0           ! damage intercept
    AA2 = 0.0028388d0     ! damage quadratic term
    AA3 = 2.0d0           ! damage exponent
    theta3 = 0.0d0        ! damage from MOC collapse
    theta4 = 0.0d0        ! damage from WEIS collapse

!  Economic parameters:
    ELAS_MU = 1.1d0       ! Elasticity of marginal utility of consumption
    GAMA = 0.3d0          ! Capital elasticity in production function
    K0 = 137.0d0          ! 2005 value of capital [trillions 2006 US$]
    DK = 0.1d0            ! Depreciation rate on capital per year
    DR = 0.03d0           ! Discount Rate from v7 GAMS

! Objective function scaling coefficients:
    scale1 = 194.0d0
    scale2 = 381800.0d0

    RETURN

END SUBROUTINE dice_params
!-----------------------------------------------------------------------------

!-----------------------------------------------------------------------------
SUBROUTINE dice_exog_var()
!  ==========================================================================
! |  Set up exogenously defined variables for DICE.                          |
!  ==========================================================================

    USE dice_in

    implicit none

    real(DP) :: PRSTP       ! Pure rate of social time preference [%]

    real(DP) :: POP_0       ! Initial population [millions]
    real(DP) :: G_POP_0     ! growth of population rate [per period]
    real(DP) :: pop_asym    ! Assymptotic population [millions]
    real(DP) :: G_fac_pop   ! population growth factor

    real(DP) :: DEL_A       ! Decline rate of technology per year
    real(DP) :: GA0         ! Initial growth rate of technology per period

    real(DP) :: A0          ! Initial level of total factor productivity

    real(DP) :: PART_FRAC_1     ! Fraction of emissions under ctrl regime 2005
    real(DP) :: PART_FRAC_2     ! Fraction of emissions under ctrl regime 2015
    real(DP) :: PART_FRAC_21    ! Fraction of emissions under ctrl regime 2205
    real(DP) :: DPART_FRAC      ! Decline rate of participation

    real(DP) :: SIG0        ! CO2-equivalent emissions-GNP ratio
    real(DP) :: g_sigma     ! CO2:GDP ratio growth rate [per period]
    real(DP) :: D_SIG       ! Decline rate of decarbonization [per yr]
    real(DP) :: D_SIG_2     ! Quadratic term in decarbonization [per yr]
    real(DP) :: GSIG

    real(DP) :: PBACK       ! Cost of backstop 2005 "per 000$ per tC 2005" [?]
    real(DP) :: BACKRAT     ! Ratio of initial to final backstop cost
    real(DP) :: GBACK       ! Initial cost decline backstop [% per period]

    real(DP) :: ELAND0      ! Carbon emissions from land [GtC per period]
    real(DP) :: ETREE_DR    ! Carbon emission from deforestation decline rate

    real(DP) :: FEX0        ! Est. of 2000 forcings of non-CO2 GHGs [W/m^2]
    real(DP) :: FEX1        ! Est. of 2100 forcings of non-CO2 GHGs [W/m^2]
    real(DP) :: F_break_yr
    real(DP) :: F_br_indx
    real(DP) :: F_slope

    integer(i4b) :: tx

    integer(i4b) :: exogIOU = 5
    integer(i4b) :: istat

! Time in years
    time_yrs = (/ ((tx - 1) * dt_in + start_year_in, tx = 1, ns) /)

! Social time preference discount factor:
    PRSTP = 0.05d0
    R(1) = PRSTP                                       ! v7 GAMS (10-yr step)
    RR(1) = 1.0d0
    do tx = 2,ns
!        R(tx) = PRSTP * exp(-0.3d0*dble(tx-1))
        R(tx) = PRSTP * exp(-DR*dt_in*dble(tx-1))
!        R(tx) = PRSTP * (1.0d0-real(tx)/real(ns))
!        RR(tx) = RR(tx-1)/((1+R(tx-1))**10)
        RR(tx) = RR(tx-1)/((1+R(tx-1))**dt_in)
    end do
!    RR = (/ ((1.0d0 + PRSTP)**(-dt_in*(tx-1)), tx = 1, ns) /)  ! v8 GAMS

! Population:
    POP_0 = 6514.0d0
    G_POP_0 = 0.35d0 / 10.0d0 * dt_in
    pop_asym = 8600.0d0

    L(1) = POP_0
    do tx = 2,ns
        G_fac_pop = (dexp(G_POP_0*(tx-1)) - 1.0d0) / dexp(G_POP_0*(tx-1))
        L(tx) = POP_0 * (1.0d0 - G_fac_pop) + G_fac_pop * pop_asym
    end do

! Growth rate of productivity:
    DEL_A = 0.001d0
    GA0 = 0.092d0 / 10.0d0 * dt_in

    GA = (/ ((GA0*dexp(-DEL_A*dt_in*(tx-1))), tx = 1, ns) /)

! Level of total factor productivity:
    A0 = 0.02722

    AL(1) = A0
    do tx = 2,ns
        AL(tx) = AL(tx-1) / (1.0d0 - GA(tx-1))
    end do

! Participation:
    PART_FRAC_1 = 1.0d0
    PART_FRAC_2 = 1.0d0
    PART_FRAC_21 = 1.0d0
    DPART_FRAC = 0.2028d0

    do tx = 1,ns
        if (time_yrs(tx) < 2015) then
            PART_FRAC(tx) = PART_FRAC_1
        elseif (time_yrs(tx) < 2205) then
            PART_FRAC(tx) = PART_FRAC_21 + (PART_FRAC_2 - PART_FRAC_21) * &
                            dexp(-DPART_FRAC*(tx-2))
        else
            PART_FRAC(tx) = PART_FRAC_21
        end if
    end do

! Emissions:
    SIG0 = 0.13418d0
    g_sigma = g_sigma_in / 10.0d0 * dt_in
    D_SIG = 0.003d0
    D_SIG_2 = 0.0d0

    SIGMA(1) = SIG0
    do tx = 2,ns
        GSIG = g_sigma * dexp(-dt_in*(D_SIG*(tx-1) + D_SIG_2*((tx-1)**2)))
        SIGMA(tx) = SIGMA(tx-1)/(1.0d0 - GSIG)
    end do

! Abatement cost:
    b2 = 2.8d0
    PBACK = 1.17d0
    BACKRAT = 2.0d0
    GBACK = 0.05d0 / 10.0d0 * dt_in

    do tx = 1, ns
        COST_1(tx) = PBACK*SIGMA(tx)/b2 * &
                    (BACKRAT - 1.0d0 + dexp(-GBACK*(tx-1)))/BACKRAT
    end do

! Carbon emissions from deforestation:
    ELAND0 = 11.0d0 / 10.0d0 * dt_in
    ETREE_DR = 0.1d0 / 10.0d0 * dt_in

    E_TREE = (/ ( ELAND0*(1.0d0 - ETREE_DR)**(tx-1), tx = 1, ns) /)

! Exogenous radiative forcing [W/m^2 since 1900]
    FEX0 = -0.06d0
    FEX1 = 0.3d0
    F_break_yr = 2100
    F_br_indx = ceiling(1.0 + real(F_break_yr - start_year_in)/dt_in)
    F_slope = 1.0 / real(F_br_indx - 1)

    do tx = 1, F_br_indx
        FORCOTH(tx) = FEX0 + F_slope*(FEX1-FEX0)*(tx-1)
    end do
    do tx = F_br_indx+1, ns
        FORCOTH(tx) = FEX1
    end do

!! Write out exog variables.
!    open(UNIT=exogIOU, FILE='exog.out', STATUS='REPLACE', ACTION='WRITE', &
!            ACCESS='SEQUENTIAL',IOSTAT=istat)
!    do tx = 1, ns
!        write(exogIOU,"(F8.2,9F15.7)") time_yrs(tx), RR(tx), L(tx), &
!            GA(tx), AL(tx), PART_FRAC(tx), SIGMA(tx), COST_1(tx), E_TREE(tx), &
!            FORCOTH(tx)
!    end do
!    close(exogIOU)

    RETURN

END SUBROUTINE dice_exog_var
!-----------------------------------------------------------------------------

!-----------------------------------------------------------------------------
SUBROUTINE dice_geophys_params()
!  ==========================================================================
! |  Pass input parameters and initialize the appropriate geophysical model. |
!  ==========================================================================

    USE dice_in

    implicit none

    call setup_DICEgeo(lambda_star_in)

    RETURN

END SUBROUTINE dice_geophys_params
!-----------------------------------------------------------------------------

!-----------------------------------------------------------------------------
SUBROUTINE dice_model(UTILITY, mu_in, IoY_in)
!  ==========================================================================
! |  Definitions of endogenous DICE quantities.                              | 
!  ==========================================================================

    USE dice_in

    implicit none

    real(DP), intent(OUT) :: UTILITY
    real(DP), dimension(ns), intent(IN) :: mu_in, IoY_in
    integer(i4b) :: tx
    real(DP) :: T_ATMOS
    real(DP) :: MOCstr

! Set initial value of DICE variables.

    call DICEgeo_initval(FORCOTH(1), T_ATMOS)


    b1(1) = PART_FRAC(1)**(1.0d0 - b2) * COST_1(1)
    frac_abate_cost(1) = b1(1)*mu_in(1)**b2
    if (damages_on) then
        frac_damages(1) = AA1*T_ATMOS + AA2*T_ATMOS**AA3
    else
        frac_damages(1) = 0.0d0
    end if

    K(1) = K0
    YGROSS(1) = AL(1)*(L(1)**(1.0d0-GAMA))*K(1)**GAMA
    Y(1) = YGROSS(1) * (1.0d0-frac_abate_cost(1)) / (1.0d0+frac_damages(1))
    production(1) = (1.0d0-mu_in(1)) * dt_in * SIGMA(1)*YGROSS(1)
    E(1) = production(1) + E_TREE(1)
    I(1) = IoY_in(1)*Y(1)
    C(1) = Y(1) - I(1)
    PERIOD_U(1) = ((C(1)/L(1))**(1.0d0-ELAS_MU) - 1.0d0) / (1.0d0 - ELAS_MU)

! Step through the rest of the time series.

    do tx = 2,ns

        call DICEgeo_step(tx, E(tx-1), FORCOTH(tx), T_ATMOS)
            
        if (T_ATMOS < 0d0) then
            UTILITY = 1000d0*dble(tx)
            RETURN
        end if

        b1(tx) = PART_FRAC(tx)**(1.0d0 - b2) * COST_1(tx)
        frac_abate_cost(tx) = b1(tx)*mu_in(tx)**b2
        if (damages_on) then
            frac_damages(tx) = AA1*T_ATMOS + AA2*T_ATMOS**AA3
        else
            frac_damages(tx) = 0.0d0
        end if

        K(tx) = (1.0d0-DK)**dt_in * K(tx-1) + dt_in * I(tx-1)
        YGROSS(tx) = AL(tx) * (L(tx)**(1.0d0-GAMA)) * K(tx)**GAMA
        Y(tx) = YGROSS(tx) * (1.0d0-frac_abate_cost(tx)) &
                    / (1.0d0+frac_damages(tx)) 
        production(tx) = (1.0d0-mu_in(tx)) * dt_in * SIGMA(tx)*YGROSS(tx)
        E(tx) = production(tx) + E_TREE(tx) 
        I(tx) = IoY_in(tx) * Y(tx)
        C(tx) = Y(tx) - I(tx)
        PERIOD_U(tx) = ((C(tx)/L(tx))**(1.0d0-ELAS_MU)-1.0d0) / (1.0d0-ELAS_MU)
    end do

    UTILITY = SUM(RR*L*PERIOD_U)*dt_in/scale1 + scale2

    RETURN

END SUBROUTINE dice_model
!-----------------------------------------------------------------------------

!!-----------------------------------------------------------------------------
!SUBROUTINE dice_out(mu_in, IoY_in)
!!  ==========================================================================
!! |  Write variables in columns to a text file in gnuplot format.
!!  ==========================================================================
!    
!    implicit none
!!
!    real(DP), dimension(ns), intent(IN) :: mu_in, IoY_in
!    integer(i4b), parameter :: doutIOU = 30
!    integer(i4b) :: tx
!    character(len=8) :: fname = 'dice.out'
!    integer(i4b) :: istat
!
!    open (doutIOU,FILE=fname,STATUS='REPLACE',ACTION='WRITE',FORM='FORMATTED', &
!            IOSTAT=istat)
!
!    do tx = 1,ns
!
!        write (doutIOU,'(f9.2,16es12.4)') time_yrs(tx), RR(tx), L(tx), GA(tx), &
!                AL(tx), SIGMA(tx), COST_1(tx), E_TREE(tx), FORCOTH(tx), Y(tx), &
!                K(tx), frac_damages(tx), E(tx), I(tx), mu_in(tx), IoY_in(tx)
!    end do
!
!    close(doutIOU)
!
!    call DICEgeo_out(time_yrs)
!
!    RETURN
!
!END SUBROUTINE dice_out
!!-----------------------------------------------------------------------------

!-----------------------------------------------------------------------------
SUBROUTINE alloc_dice()
!  =========================================================================
! |  Allocate dice model variables.                                         |
!  =========================================================================

    implicit none

    integer(i4b) :: astat

    allocate(time_yrs(ns),STAT=astat)
    allocate(R(ns),STAT=astat)
    allocate(RR(ns),STAT=astat)
    allocate(L(ns),STAT=astat)
    allocate(GA(ns),STAT=astat)
    allocate(AL(ns),STAT=astat)
    allocate(PART_FRAC(ns),STAT=astat)
    allocate(SIGMA(ns),STAT=astat)
    allocate(COST_1(ns),STAT=astat)
    allocate(E_TREE(ns),STAT=astat)
    allocate(FORCOTH(ns),STAT=astat)
    allocate(b1(ns),STAT=astat)
    allocate(frac_abate_cost(ns),STAT=astat)
    allocate(frac_damages(ns),STAT=astat)
    allocate(K(ns),STAT=astat)
    allocate(YGROSS(ns),STAT=astat)
    allocate(Y(ns),STAT=astat)
    allocate(production(ns),STAT=astat)
    allocate(E(ns),STAT=astat)
    allocate(I(ns),STAT=astat)
    allocate(C(ns),STAT=astat)
    allocate(PERIOD_U(ns),STAT=astat)

    RETURN

END SUBROUTINE alloc_dice
!-----------------------------------------------------------------------------

!-----------------------------------------------------------------------------
SUBROUTINE dealloc_dice()
!  =========================================================================
! |  Deallocate dice model and geophysical variables.                       |
!  =========================================================================

    implicit none

    integer(i4b) :: astat

    astat = 0

    deallocate(time_yrs,STAT=astat)
    deallocate(R,STAT=astat)
    deallocate(RR,STAT=astat)
    deallocate(L,STAT=astat)
    deallocate(GA,STAT=astat)
    deallocate(AL,STAT=astat)
    deallocate(PART_FRAC,STAT=astat)
    deallocate(SIGMA,STAT=astat)
    deallocate(COST_1,STAT=astat)
    deallocate(E_TREE,STAT=astat)
    deallocate(FORCOTH,STAT=astat)
    deallocate(b1,STAT=astat)
    deallocate(frac_abate_cost,STAT=astat)
    deallocate(frac_damages,STAT=astat)
    deallocate(K,STAT=astat)
    deallocate(YGROSS,STAT=astat)
    deallocate(Y,STAT=astat)
    deallocate(production,STAT=astat)
    deallocate(E,STAT=astat)
    deallocate(I,STAT=astat)
    deallocate(C,STAT=astat)
    deallocate(PERIOD_U,STAT=astat)

    call dealloc_DICEgeo()

    if (astat > 0) print *, "Problem deallocating dice_dt arrays"

    RETURN

END SUBROUTINE dealloc_dice
!-----------------------------------------------------------------------------
END MODULE DICE
