!    DICE_GEO:  Geophysical model for Nordhaus_DICE
!    Copyright (C) 2010  Klaus Keller and Brian Tuttle 
!
!    This program is free software; you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation; either version 2 of the License, or
!    (at your option) any later version.
!
!    This program is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with this program; if not, write to the Free Software
!    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
!
!    Klaus Keller, kkeller@geosc.psu.edu
!    Brian Tuttle, btuttle@psu.edu
!
!-----------------------------------------------------------------------------
MODULE dice_geo
    
    USE global

    implicit none

    private

    real(DP) :: dt
    integer(i4b) :: ns

! Climate damage parameters:
    real(DP) :: AA1, AA2, AA3, theta3, theta4

! Radiative forcing parameters:
    real(DP) :: FCO22X, preanthCO2

! Climate parameters:
    real(DP) :: T_ATM_0, C1, C3, lambda_star, LAM, T_OCEAN0, C4

! Carbon cycle parameters:
    real(DP) :: M_AT_2000, M_UP_2000, M_LO_2000
    real(DP) :: b11, b12, b21, b22, b23, b32, b33

    real(DP), dimension(:), allocatable, public :: FORC
    real(DP), dimension(:), allocatable, public :: T_ATM
    real(DP), dimension(:), allocatable, public :: M_AT
    real(DP), dimension(:), allocatable :: T_OCEAN
    real(DP), dimension(:), allocatable :: M_UP, M_LO

    public :: init_DICEgeo_arrays, setup_DICEgeo, DICEgeo_initval
    public :: DICEgeo_step, dealloc_DICEgeo, DICEgeo_out

CONTAINS
!-----------------------------------------------------------------------------
SUBROUTINE init_DICEgeo_arrays(nsteps, stepsize)
!  ==========================================================================
! |  
!  ==========================================================================

    implicit none

    integer(i4b), intent(IN) :: nsteps
    real(DP), intent(IN) :: stepsize

    ns = nsteps
    dt = stepsize

    call alloc_DICEgeo()

    RETURN

END SUBROUTINE init_DICEgeo_arrays
!-----------------------------------------------------------------------------

!-----------------------------------------------------------------------------
SUBROUTINE setup_DICEgeo(lambda_star)
!  ==========================================================================
! | Inputs:
! |  lambda_star: Climate sens, change in T_atm per CO2 doubling [K]  3.6 def  
!  ==========================================================================

    implicit none

    real(DP), intent(IN) :: lambda_star

! Radiative forcing parameters:
    FCO22X = 3.7d0        ! Radiative forcing of eq. CO2 doubling [W/m^2]
    preanthCO2 = 596.4d0  ! Pre-anthopogenic CO2 [GtC]

! Climate parameters:
    T_ATM_0 = 0.7307d0    ! Atmospheric temperature change in 2000 from 1900 [K]
    C1 = 0.22d0 / 10.0d0 * dt   ! Climate eq. coef. for upper level [per period]
    C3 = 0.3d0            ! Transfer coefficient upper to lower stratum
    LAM = FCO22X/lambda_star
    T_OCEAN0 = 0.0068d0   ! Lower strat. temp. change in 2000 from 1900 [K]
    C4 = 0.05d0 / 10.0d0 * dt   ! Transfer coef. for lower level [per period]

! Carbon cycle parameters:
    M_AT_2000 = 808.9d0     ! Concentration in atmosphere 2005 [GtC]
    M_UP_2000 = 1255.0d0    ! Concentration in upper strata 2005 [GtC]
    M_LO_2000 = 18365.0d0   ! Concentration in lower strata 2000 [GtC]
! Carbon cycle transition matrix:
    b11 = 0.810712d0**(dt/10.0d0)
    b12 = 0.189288d0 / 10.0d0 * dt
    b21 = 0.097213d0 / 10.0d0 * dt
    b22 = 0.852787d0**(dt/10.0d0)
    b23 = 0.05d0 / 10.0d0 * dt
    b32 = 0.003119d0 / 10.0d0 * dt
    b33 = 0.996881d0**(dt/10.0d0)

    RETURN

END SUBROUTINE setup_DICEgeo
!-----------------------------------------------------------------------------

!-----------------------------------------------------------------------------
SUBROUTINE DICEgeo_initval(Fex,T_at)
!  ==========================================================================
! |  Define initial values at time step 1.
!  ==========================================================================

    implicit none

    real(DP), intent(IN) :: Fex
    real(DP), intent(OUT) :: T_at

    M_AT(1) = M_AT_2000
    M_UP(1) = M_UP_2000
    M_LO(1) = M_LO_2000

    FORC(1) = FCO22X*(dlog((M_AT(1)+1.0d-6)/preanthCO2)/dlog(2.0d0)) + Fex

    T_ATM(1) = T_ATM_0
    T_OCEAN(1) = T_OCEAN0

    T_at = T_ATM(1)

    RETURN

END SUBROUTINE DICEgeo_initval
!-----------------------------------------------------------------------------

!-----------------------------------------------------------------------------
SUBROUTINE DICEgeo_step(tx, Emissions, Fex, T_at)
!  ==========================================================================
! |  Propagate the model one time step.
!  ==========================================================================

    implicit none

    integer(i4b), intent(IN) :: tx
    real(DP), intent(IN) :: Emissions
    real(DP), intent(IN) :: Fex
    real(DP), intent(OUT) :: T_at

!    M_AT(tx) = E(tx-1) + b11*M_AT(tx-1) + b21*M_UP(tx-1)

    M_AT(tx) = Emissions + b11*M_AT(tx-1) + b21*M_UP(tx-1)
    M_UP(tx) = b12*M_AT(tx-1) + b22*M_UP(tx-1) + b32*M_LO(tx-1)
    M_LO(tx) = b33*M_LO(tx-1) + b23*M_UP(tx-1)

    FORC(tx) = FCO22X*(dlog((M_AT(tx)+1.0d-6)/preanthCO2)/dlog(2.0d0)) + Fex

    T_ATM(tx) = T_ATM(tx-1) + &
        C1*(FORC(tx)-LAM*T_ATM(tx-1)-C3*(T_ATM(tx-1)-T_OCEAN(tx-1)))
    T_OCEAN(tx) = T_OCEAN(tx-1) + C4*(T_ATM(tx-1)-T_OCEAN(tx-1))

! Assign output.
    T_at = T_ATM(tx)

    RETURN

END SUBROUTINE DICEgeo_step
!-----------------------------------------------------------------------------

!-----------------------------------------------------------------------------
SUBROUTINE DICEgeo_out(yrs) 
    
    implicit none

    real(SP), dimension(:), intent(IN) :: yrs
    integer(i4b), parameter :: dgeoIOU = 50
    integer(i4b) :: tx
    character(len=12) :: fname = 'dice_geo.out'
    integer(i4b) :: istat

    open(dgeoIOU,FILE=fname,STATUS='REPLACE',ACTION='WRITE',FORM='FORMATTED', &
            IOSTAT=istat)
    do tx = 1,ns
        write(dgeoIOU,'(f9.2,6es12.4)') yrs(tx), T_ATM(tx), T_OCEAN(tx), &
                                        FORC(tx), M_AT(tx), M_UP(tx), M_LO(tx)
    end do

    close(dgeoIOU)

    RETURN

END SUBROUTINE DICEgeo_out
!-----------------------------------------------------------------------------

!-----------------------------------------------------------------------------
SUBROUTINE alloc_DICEgeo()

    implicit none

    integer(i4b) :: astat

    allocate(M_AT(ns),STAT=astat)
    allocate(M_UP(ns),STAT=astat)
    allocate(M_LO(ns),STAT=astat)

    allocate(FORC(ns),STAT=astat)

    allocate(T_ATM(ns),STAT=astat)
    allocate(T_OCEAN(ns),STAT=astat)

    RETURN

END SUBROUTINE alloc_DICEgeo
!-----------------------------------------------------------------------------

!-----------------------------------------------------------------------------
SUBROUTINE dealloc_DICEgeo()

    implicit none

    integer(i4b) :: astat

    deallocate(M_AT,STAT=astat)
    deallocate(M_UP,STAT=astat)
    deallocate(M_LO,STAT=astat)

    deallocate(FORC,STAT=astat)

    deallocate(T_ATM,STAT=astat)
    deallocate(T_OCEAN,STAT=astat)

    RETURN

END SUBROUTINE dealloc_DICEgeo
!-----------------------------------------------------------------------------
END MODULE dice_geo
