!    DICE_in: Input module for DICE07 with variable time step
!    Copyright (C) 2010  Klaus Keller and Brian Tuttle 
!
!    This program is free software; you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation; either version 2 of the License, or
!    (at your option) any later version.
!
!    This program is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with this program; if not, write to the Free Software
!    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
!
!    Klaus Keller, klaus.psu@gmail.com
!    Brian Tuttle, btuttle@psu.edu
!

!-----------------------------------------------------------------------------
MODULE dice_in

    USE global

    implicit none

!  ==========================================================================
! |  Input variables and their initial default values are global to this
! |  module.  The defaults may be overwritten by the read_dice_input() 
! |  subroutine.
!  ==========================================================================

! DICE problem size:
    integer(i4b) :: start_year_in           = 2005
    integer(i4b) :: end_year_in             = 2595
    real(DP) :: dt_in                       = 1.0d0

! Optimization problem size:
!    character (len=20) :: opt_type
    character (len=20) :: opt_type_in       = 'abate'
    integer(i4b) :: param_start_in          = 2005
    integer(i4b) :: param_end_in            = 2595
    character (len=20) :: time_step_in      = '10_year'

! Climate sensitivity:
    real(DP) :: lambda_star_in              = 3.0d0

! CO2:GDP growth rate:
    real(DP) :: g_sigma_in                  = -0.073d0

CONTAINS
!-----------------------------------------------------------------------------
SUBROUTINE read_dice_input(IOU, errcode)
!  ==========================================================================
! |  Read the portion of the input file pertaining to dice07 model vars.     |
!  ==========================================================================

    implicit none

    integer(i4b), intent(IN) :: IOU
    integer(i4b), intent(OUT) :: errcode
    integer, parameter :: tabstop = 31
    character(len=tabstop) :: tabover
    character(len=75) :: sec_head
    character(len=1) :: numstr
    integer :: istat, i
    integer(i4b) :: stat = 0

  if (iam == 0) then

    headers: do
      read(IOU, "(a)", IOSTAT=istat) sec_head

      select case (trim(sec_head(1:29)))
      case ('Optimization variable [abate,')
        opt_vars: do
           read(IOU, "(a)", ADVANCE="NO", IOSTAT=istat) tabover
           select case (tabover)
           case ('  optimization_type          = ')
              read(IOU, "(a)", IOSTAT=istat) opt_type_in
           case ('                               ')
              exit opt_vars
           end select
        end do opt_vars

      case ('Optimization parameters:')
        opt_param: do
           read(IOU, "(a)", ADVANCE="NO", IOSTAT=istat) tabover
           select case (tabover)
           case ('  Parameter start [year]     = ')
              read(IOU, "(i4)", IOSTAT=istat) param_start_in
           case ('  Parameter end [year]       = ')
              read(IOU, "(i4)", IOSTAT=istat) param_end_in
           case ('                               ')
              exit opt_param
           end select
        end do opt_param

      case ('Parameter interval [10_year,')
        param_int: do
           read(IOU, "(a)", ADVANCE="NO", IOSTAT=istat) tabover
           select case (tabover)
           case ('  Time interval              = ')
              read(IOU, "(a)", IOSTAT=istat) time_step_in
           case ('                               ')
              exit param_int
           end select
        end do param_int

      case ('Model size:')
        ps_vars: do
           read(IOU, "(a)", ADVANCE="NO", IOSTAT=istat) tabover
           select case (tabover)
           case ('  Model start [year]         = ')
              read(IOU, "(i4)", IOSTAT=istat) start_year_in
           case ('  Model end [year]           = ')
              read(IOU, "(i4)", IOSTAT=istat) end_year_in
           case ('  Model time step [years]    = ')
              read(IOU, "(f7.4)", IOSTAT=istat) dt_in
           case ('                               ')
              exit ps_vars
           end select
        end do ps_vars

      case ('DICE parameters:')
        DICE_vars: do
           read(IOU, "(a)", ADVANCE="NO", IOSTAT=istat) tabover
           select case (tabover)
           case ('  Climate sensitivity        = ')
              read(IOU, "(f7.4)", IOSTAT=istat) lambda_star_in
           case ('  G_sigma (def -0.073)       = ')
              read(IOU, "(f7.4)", IOSTAT=istat) g_sigma_in
           case ('                               ')
              exit DICE_vars
           end select
        end do DICE_vars

      case ('-------------------- END MODE')
        exit headers
      end select
    end do headers

  end if

    call check_input(stat)
    errcode = errcode + stat

#ifdef HAVE_MPI
    call MPI_Bcast(start_year_in, 1, MPI_Integer, MASTER, MPI_Comm_World, ierr)
    call MPI_Bcast(end_year_in, 1, MPI_Integer, MASTER, MPI_Comm_World, ierr)
    call MPI_Bcast(dt_in, 1, MPI_Double_Precision, MASTER, MPI_Comm_World, ierr)
    call MPI_Bcast(opt_type_in, len(opt_type_in), MPI_Character, MASTER, MPI_Comm_World, ierr)
    call MPI_Bcast(param_start_in, 1, MPI_Integer, MASTER, MPI_Comm_World, ierr)
    call MPI_Bcast(param_end_in, 1, MPI_Integer, MASTER, MPI_Comm_World, ierr)
    call MPI_Bcast(time_step_in, len(time_step_in), MPI_Character, MASTER, MPI_Comm_World, ierr)
    call MPI_Bcast(lambda_star_in, 1, MPI_Double_Precision, MASTER, MPI_Comm_World, ierr)
    call MPI_Bcast(g_sigma_in, 1, MPI_Double_Precision, MASTER, MPI_Comm_World, ierr)
#endif


    RETURN

END SUBROUTINE read_dice_input
!-----------------------------------------------------------------------------

!-----------------------------------------------------------------------------
SUBROUTINE check_input(errcode)
!  ==========================================================================
! | Checks input file for inconsistencies.
!  ==========================================================================

    implicit none

    integer(i4b), intent(OUT) :: errcode

    errcode = 0

    if (start_year_in < param_start_in) then
        print *, "Model start year must be >= starting parameter."
        errcode = 1
    end if

    if (end_year_in > param_end_in) then
        print *, "Model end year must be <= ending parameter."
        errcode = 1
    end if

    RETURN

END SUBROUTINE check_input
!-----------------------------------------------------------------------------
END MODULE dice_in
