!    DICE_OBJFN: Interface module for DICE with optimization software
!    Copyright (C) 2010  Klaus Keller and Brian Tuttle 
!
!    This program is free software; you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation; either version 2 of the License, or
!    (at your option) any later version.
!
!    This program is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with this program; if not, write to the Free Software
!    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
!
!    Klaus Keller, klaus.psu@gmail.com
!    Brian Tuttle, btuttle@psu.edu
!
!-----------------------------------------------------------------------------
MODULE dice_objfn

    USE global
    USE dice
    USE dice_in

    implicit none

    integer(i4b), target :: nop         ! number of model parameters
    integer(i4b), target :: ng          ! number of constraints

! Corresponding time series for optimization paramer set:
    integer(i4b), dimension(:), allocatable :: time_param

! Business-as-usual investment solution for use with abatement optimization:
    real(DP), dimension(:), allocatable :: IoY_BAU
    real(DP), dimension(:), allocatable :: IoY_arr
    real(DP), dimension(:), allocatable :: mu_arr

! Break points for adaptive parameter time series:
    integer(i4b) :: adaptBP1 = 2165
    integer(i4b) :: adaptBP2 = 2285

CONTAINS
!-----------------------------------------------------------------------------
SUBROUTINE model_size(errcode)
!  ==========================================================================
! |  Report size of the DICE model for use with optimization software.       |
!  ==========================================================================

    implicit none

    integer(i4b), intent(OUT) :: errcode

    errcode = 0

! Find the size of the dice model with variable time step.
    call dice_size()

! Number of constraints:
    ng = 1

! Find the number of parameters for optimization:  Rounding to the next
! highest integer allows for partial time periods, and adding one includes
! the starting year.
    if (time_step_in == '10_year') then
        nop = ceiling((param_end_in - param_start_in)/10.0) + 1

    else if (time_step_in == '25_year' ) then
        nop = ceiling((param_end_in - param_start_in)/25.0) + 1

    else if (time_step_in == 'adaptive' ) then
        if (param_start_in >= adaptBP1) then
            print *, "Starting parameter too high for adaptive time step."
            errcode = 1
            RETURN
        end if
        if (param_end_in <= adaptBP1) then
            nop = ceiling((param_end_in - param_start_in)/10.0) + 1
        elseif (param_end_in <= adaptBP2) then
            nop = ceiling((adaptBP1 - param_start_in)/10.0) + &
                 ceiling((param_end_in - adaptBP1)/20.0) + 1
        else
            nop = ceiling((adaptBP1 - param_start_in)/10.0) + &
                 int((adaptBP2 - adaptBP1)/20.0) + &
                 ceiling((param_end_in - adaptBP2)/40.0) + 1
        end if
    else
        print *, "Invalid argument for time_step: ", time_step_in
        errcode = 1
        RETURN
    end if

    call alloc_dice_objf()

    RETURN

END SUBROUTINE model_size
!-----------------------------------------------------------------------------

!-----------------------------------------------------------------------------
SUBROUTINE model_setup(lower_limit, upper_limit, errcode)
!  ==========================================================================
! | Setup the DICE model for use with optimization software.                 |
!  ==========================================================================

    implicit none

    real(DP), dimension(:), intent(OUT) :: lower_limit, upper_limit
    integer(i4b), intent(OUT) :: errcode
    integer(i4b) :: i
    integer(i4b) :: adapt_int = 10
    integer(i4b) :: errstat = 0

! Define upper and lower bounds of parameter space.
    lower_limit = 0.0d0
    upper_limit = 1.0d0

! Define the time series corresponding to the optimization parameter set.
    if (time_step_in == '10_year') then
        time_param = (/ (param_start_in + 10*(i-1), i=1,nop) /)

    else if (time_step_in == '25_year' ) then
        time_param = (/ (param_start_in + 25*(i-1), i=1,nop) /)

    else if (time_step_in == 'adaptive' ) then
        time_param(1) = param_start_in
        do i = 2,nop
            if (time_param(i-1) >= adaptBP1) adapt_int = 20
            if (time_param(i-1) >= adaptBP2) adapt_int = 40
            time_param(i) = time_param(i-1) + adapt_int
        end do
    end if
    
! Define investment time series for abatement optimization.
    select case (opt_type_in)
    case ('BAU_invest')
        damages_on = .false.
        mu_arr = 0d0
    case ('abate')
        damages_on = .true.
        call load_IoY(errstat)
        if (errstat == 0) IoY_arr = IoY_BAU
    end select

! Initialize DICE model.
    call init_dice()

    errcode = errstat

    RETURN

END SUBROUTINE model_setup
!-----------------------------------------------------------------------------

!-----------------------------------------------------------------------------
SUBROUTINE model_out(soln, obf)
!  ==========================================================================
! |  Run the DICE model and write out the state of model variables to files. |
!  ==========================================================================

    USE dice_out

    implicit none

    real(DP), dimension(nop), intent(IN) :: soln
    real(DP), intent(OUT) :: obf
    real(DP), dimension(ng) :: constr
    integer(i4b) :: cnt

! Run the model once.
    call calc_fg_dice(soln, obf, constr, cnt)

! Write out model variables.
    call write_dice(mu_arr, IoY_arr)

    RETURN

END SUBROUTINE model_out
!-----------------------------------------------------------------------------

!-----------------------------------------------------------------------------
SUBROUTINE calc_fg_dice(x, f, g, counter)
!  ==========================================================================
! | Objective function calls DICE model, calculates penalties, counts model  |
! | evaluations.                                                             |
!  ==========================================================================

    implicit none

    real(DP), dimension(nop), intent(IN) :: x
    real(DP), intent(OUT) :: f
    real(DP), dimension(ng), intent(OUT) :: g
    integer(i4b), intent(OUT) :: counter
    real(DP) :: util, penalty

! Initialize cumulative variables.
    penalty = 0d0
    counter = 0

! Build time series from set of optimization parameters x.
    select case (opt_type_in)
    case ('BAU_invest')

        IoY_arr = linterp_param(x)

    case ('abate') 

        mu_arr = linterp_param(x)

    end select

    call dice_model(util, mu_arr, IoY_arr)
    counter = counter + 1


! Define objective function value.
    f = -util

! Assign penalties.
    f = f + penalty

! Assign constraints.
    g(1) = 0d0

    RETURN

END SUBROUTINE calc_fg_dice
!-----------------------------------------------------------------------------

!-----------------------------------------------------------------------------
FUNCTION linterp_param(coarse_arr) RESULT(fine_arr)
!  ==========================================================================
! |  Linear interpolation of a coarse array of optimization parameters
! |  to a fine array of a DICE time-series model variable with reference to
! |  the corresponding time arrays.
!  ==========================================================================

    implicit none

    real(DP), dimension(nop) :: coarse_arr
    real(DP), dimension(ns) :: fine_arr
    real(DP) :: slope
    integer(i4b) :: i, j

    j = 1
    do i = 1,nop-1
        if (time_yrs(j) > time_param(i+1)) cycle
        slope = (coarse_arr(i+1) - coarse_arr(i)) / &
                (time_param(i+1) - time_param(i))
        do
            fine_arr(j) = coarse_arr(i) + slope*(time_yrs(j)-time_param(i))
            j = j + 1
            if (j > ns) exit
            if (time_yrs(j) >= time_param(i+1)) exit
        end do
    end do
    if (time_yrs(ns) == time_param(nop)) fine_arr(ns) = coarse_arr(nop)


END FUNCTION linterp_param
!-----------------------------------------------------------------------------

!-----------------------------------------------------------------------------
SUBROUTINE load_IoY(errcode)

    implicit none

    integer(i4b), intent(OUT) :: errcode
    integer(i4b), parameter :: end_of_file = -1
    integer(i4b), parameter :: invIU = 10
    character(len=7) :: fname = 'IoY.bau'
    logical :: theresafile
    integer(i4b) :: i, istat, IoYsize
    integer(i4b) :: errstat = 0

#ifdef HAVE_MPI
  if (iam == MASTER) then
#endif

! Check for file.
    inquire(file=fname,exist=theresafile)

    if (theresafile) then

! Check size of file.
        IoYsize = 0
        open(invIU,FILE=fname,STATUS='OLD',ACTION='READ',FORM='FORMATTED', &
                IOSTAT=istat)
        do
            read(invIU,"(F15.7)",IOSTAT=istat) IoY_BAU(1)
            if (istat == end_of_file) then
                exit
            else
                IoYsize = IoYsize + 1
            end if
        end do

        close(invIU)

        if (IoYsize == ns) then
! Read file.
            open(invIU,FILE=fname,STATUS='OLD',ACTION='READ',FORM='FORMATTED', &
                    IOSTAT=istat)
            do i = 1,ns
                read(invIU,"(F15.7)",IOSTAT=istat) IoY_BAU(i)
            end do

            close(invIU)

        else
            print *, 'load_IoY: Size of data file IoY.bau', IoYsize, 'does not match allocated memory', ns
            errstat = 1
        end if
    else
        print *, 'load_IoY: IoY.bau file not found.'
        errstat = 1
    end if

#ifdef HAVE_MPI
  end if

! Share with other procs.
    call MPI_Bcast(IoY_BAU, ns, MPI_Double_Precision, MASTER, MPI_Comm_World, ierr)
    call MPI_Bcast(errstat, 1, MPI_Integer, MASTER, MPI_Comm_World, ierr)
#endif
    
    errcode = errstat

    RETURN

END SUBROUTINE load_IoY
!-----------------------------------------------------------------------------

!-----------------------------------------------------------------------------
SUBROUTINE print_model_input()

    implicit none

    print *, "MODEL PARAMETERS:"
    print *, "Optimization variable [abate, BAU_invest]:"
    print *, "  optimization_type          = ", opt_type_in
    print *, ""
    print *, "Optimization parameters:"
    print *, "  Parameter start [year]     = ", param_start_in
    print *, "  Parameter end [year]       = ", param_end_in
    print *, ""
    print *, "Parameter interval [10_year, 25_year, adaptive]"
    print *, "  Time interval              = ", time_step_in
    print *, ""
    print *, "Model size:"
    print *, "  Model start [year]         = ", start_year_in
    print *, "  Model end [year]           = ", end_year_in
    print *, "  Model time step [years]    = ", dt_in
    print *, ""
    print *, "DICE parameters:"
    print *, "  Climate sensitivity        = ", lambda_star_in
    print *, "  G_sigma (def -0.073)       = ", g_sigma_in


    RETURN

END SUBROUTINE print_model_input
!-----------------------------------------------------------------------------

!-----------------------------------------------------------------------------
SUBROUTINE alloc_dice_objf()

    implicit none

    integer(i4b) :: astat

    allocate(time_param(nop), STAT=astat)
    allocate(mu_arr(ns), STAT=astat)
    allocate(IoY_arr(ns), STAT=astat)

    if (opt_type_in .ne. "BAU_invest") then
        allocate(IoY_BAU(ns), STAT=astat)
    end if

    RETURN

END SUBROUTINE alloc_dice_objf
!-----------------------------------------------------------------------------

!-----------------------------------------------------------------------------
SUBROUTINE dealloc_dice_objf()

    implicit none

    integer(i4b) :: astat

    deallocate(time_param, STAT=astat)
    deallocate(mu_arr, STAT=astat)
    deallocate(IoY_arr, STAT=astat)
    if (allocated(IoY_BAU)) deallocate(IoY_BAU, STAT=astat)

    call dealloc_dice()

    RETURN

END SUBROUTINE dealloc_dice_objf
!-----------------------------------------------------------------------------
END MODULE dice_objfn
